/*
 * Causal Path Discovery for each request
 * author : Anup Goyal
 *
 * goal: Parse the network traces of coral web , dns and dhst traffic and store it in the 
 * causal path data structure.
 */


#include "coral_pkt_parser.h"

#define DHTIDSIZE 8 

/** BEGIN FIX **/
namespace __gnu_cxx                                                                                 
{                                                                                             
	template<> struct hash< std::string >                                                       
	{                                                                                           
		size_t operator()( const std::string& x ) const                                           
		{                                                                                         
			return hash< const char* >()( x.c_str() );                                              
		}                                                                                         
	};

}


enum pathType{DNS_REQ,DNS_RES,DNS,DHT,HTTP,HTTP_REQ,HTTP_RES};

struct mID{
	std::string urlID;
	std::string sha1ID;
};

class TimeDiff{
	public:
		timeval m_tm[4];
		int m_flowCount;
		pathType m_type;
	public:
		TimeDiff();
		double getOffsetTimeDiff();
		double getProcessingTimeDiff();
		double getOutgoingLatency();
		double getIncomingLatency();
		double getResponseTime(bool http=true);
};



/** BEGIN FIX **/
namespace __gnu_cxx 																				
{
	template<> struct hash<const TimeDiff* >                                                       
	{                                                                                           
		size_t operator()( const TimeDiff*  x ) const                                           
		{                                                                                         
			return hash< long >()( (long)x );                                              
		}                                                                                         
	};
}


class Flow{
	public:
		unsigned int m_sip;
		unsigned int m_dip;
		struct timeval m_srttime;
		std::string m_id;
		pathType m_type;
	public:
		Flow(unsigned int sip,unsigned int dip,struct timeval srttime,std::string &id,pathType type):m_sip(sip),m_dip(dip),m_srttime(srttime),m_id(id),m_type(type){}
		//~Flow();
		void getType(std::string &ptype,std::string &color);
		void printFlow(int &flnum);
		void getPrintStr(std::string &flwstr,int flnum=0,bool isdot=false);
		void getUniqueFlowIDString(std::string &fid);
		void getTimevalString(std::string &tmval);
};


class CausalPath{
	public:
		 
		std::string m_urlID;  // HTTP url Identifier
		std::string m_sha1ID; // SHA1 HASH id of the URL

		struct timeval m_startTime; // flow start time in secs.msec
		struct timeval m_endTime; // flow end time in secs.msec
		struct timeval m_curTime; // Current Packet time

		unsigned int coralDnsIp;
		unsigned int coralWebIp;
		unsigned int m_clietIP;

		hash_map<std::string,int> m_dht_ids;
		std::list <Flow *>  m_flows;

		unsigned int m_rtt;
		bool m_isMatchResponse;
		bool m_isAmbigous;
		bool m_isRetransmission;
		

	public:
		CausalPath();
		~CausalPath();
		CausalPath(std::string &sid);
		CausalPath(std::string &urlid,unsigned int &sip, unsigned int &dip, struct timeval &stime, pathType &type, std::string &httpid);
		void init(std::string &sid);
		void printPath();
		bool addFlow(struct flow);
		bool addFlow(unsigned int &sip,unsigned int &dip,struct timeval &srttime, std::string &id, pathType &type);
		std::string addDHTId(std::string &hexData);
		bool isMatchPath(std::string &hexData, unsigned int &sip, unsigned int &dip);
		bool isMatchID(std::string &urlid, unsigned int &sip, unsigned int &dip, pathType &type,bool &is_ambiguous,std::string &httpid,bool &isRetransmission, struct pcap_pkthdr& hdr);
		void removeDupFlow(hash_map<std::string,std::string> &unq_flow,bool isdot=false);
		void getIpTypeStr(hash_map<unsigned int, std::string> &ipType,Flow* &flow);
		void getProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap);
		void printProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap);
};


/** BEGIN FIX **/
namespace __gnu_cxx 																				
{
	template<> struct hash<const CausalPath * >                                                       
	{                                                                                           
		size_t operator()( const CausalPath*  x ) const                                           
		{                                                                                         
			return hash< long >()( (long)x );                                              
		}                                                                                         
	};
	template<> struct hash<const std::list<double> * >                                                       
	{                                                                                           
		size_t operator()( const std::list<double> *  x ) const                                           
		{                                                                                         
			return hash< long >()( (long)x );                                              
		}                                                                                         
	};
}


typedef hash_map<std::string, CausalPath *> path_ID_Hashmap;
typedef hash_map<std::string, CausalPath *>::iterator path_id_it;

typedef hash_map<int, CausalPath *> path_time_Hashmap;
typedef hash_map<int, CausalPath *>::iterator path_time_it;
typedef hash_map<std::string,std::list<double> *> ip_procTime_hashmap;



class Graphs{
	public:
		int m_total_path;
		int m_ambiguous_req;
		int m_retransmission_req;
		int m_tot_http_req;
		int m_tot_dns_req;
		int m_tot_dht_req;
		path_ID_Hashmap m_path_id_map;
		path_time_Hashmap m_path_time_map;
		std::list<CausalPath *> m_path_list;
		std::list<Flow *> m_dhtpkt_list;
		ip_procTime_hashmap m_ipHttpProcessingTime_map;
		ip_procTime_hashmap m_ipDhtProcessingTime_map;
		ip_procTime_hashmap m_ipDnsProcessingTime_map;


	protected:
		void parseHttpReq(std::map<std::string, std::string> &prm, std::string& hdr, std::string& path );
		void print_header(struct pcap_pkthdr& hdr,	struct miphdr* &iph);
		void hexdump(std::string &data, unsigned char *cstr, int &len);
		
		void makeConnId(std::string &connid,unsigned int &sip, unsigned short &port);
		void updatePathID(std::string &urlid, struct pcap_pkthdr& hdr,struct miphdr* &iph , pathType type,std::string &httpid,unsigned short &dp);
	public:
		Graphs();
		~Graphs();
		void parseTCPPkt(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr, struct miphdr* &iph, struct mtcphdr* &tcphdr);
		void parseUDPPkt(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr, struct miphdr* &iph, struct mudphdr* &udphdr);
		void parse_dns(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr,struct miphdr* &iph,struct mudphdr* &udphdr);
		void printGraph();
		void matchDHTpacket(std::string &hexData, struct pcap_pkthdr& hdr,struct miphdr* &iph);
		void matchHTTPRes(std::string &connid, struct pcap_pkthdr& hdr,struct miphdr* &iph);
		struct timeval time_difference(struct timeval &first, struct timeval &second);
		void updateProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap,CausalPath* &cpath);
		void printProcessingTimeStats();
		void printPTime(ip_procTime_hashmap &ipPTime_map,char *type);
		//bool checkThresTime(timeval* t1, timeval* t2, int time_thres);
		
};



