#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>

#include "hadoop.h"
#include "util.h"

bool bOldVersion = false;


int readInt(char *stream)
{
	int i = 0;
	for (int idx = 0; idx < 4; idx++) {
		char b = stream[idx];
		i = i << 8;
		i = i | (b & 0xFF);
	}
	return i;
}

long long readLong(char *stream)
{
	long long i = 0;
	for (int idx = 0; idx < 8; idx++) {
		char b = stream[idx];
		i = i << 8;
		i = i | (b & 0xFF);
	}
	return i;
}

short readShort(char *stream)
{
	short i = 0;
	for (int idx = 0; idx < 2; idx++) {
		char b = stream[idx];
		i = i << 8;
		i = i | (b & 0xFF);
	}
	return i;
}

unsigned short readUShort(char *stream)
{
	unsigned short i = 0;
	for (int idx = 0; idx < 2; idx++) {
		char b = stream[idx];
		i = i << 8;
		i = i | (b & 0xFF);
	}
	return i;
}

float readFloat(char *stream)
{
	int *pi = (int *)stream;
	int i = *pi;
	i = htonl(i);
	float *f = (float *) &i;
	return *f;
}


/**
   * Parse the first byte of a vint/vlong to determine the number of bytes
   * @param value the first byte of the vint/vlong
   * @return the total number of bytes (1 to 9)
   */
int decodeVIntSize(char value) 
{
	if (value >= -112) {
		return 1;
	} else if (value < -120) {
		return -119 - value;
	}
	return -111 - value;
}

 /**
   * Given the first byte of a vint/vlong, determine the sign
   * @param value the first byte
   * @return is the value negative
   */
bool isNegativeVInt(char value) 
{
    return value < -120 || (value >= -112 && value < 0);
}
 

/**
   * Reads a zero-compressed encoded long from input stream and returns it.
   * @param stream Binary input stream
   * @throws java.io.IOException 
   * @return deserialized long from stream.
   */
 long long readVLong(char *stream, int &len) {
	char firstByte = *stream;
	len = decodeVIntSize(firstByte);
	if (len == 1) {
		return firstByte;
	}
	long i = 0;
	for (int idx = 0; idx < len-1; idx++) {
		char b = stream[idx+1];
		i = i << 8;
		i = i | (b & 0xFF);
	}
	
	return (isNegativeVInt(firstByte) ? (i ^ -1L) : i);
}

  /**
   * Reads a zero-compressed encoded integer from input stream and returns it.
   * @param stream Binary input stream
   * @throws java.io.IOException 
   * @return deserialized integer from stream.
   */
int readVInt(char* stream, int &len) 
{
	return (int) readVLong(stream, len);
}

//////////////////////////////////////////////////////////////////////////////////
bool JStream::skip(int nbytes)
{
	if(nbytes > len_)
		return false;

	len_-= nbytes;
	data_ += nbytes;
	return true;
}

bool JStream::skip_hrpc_header()
{
	if( strncmp(data_, "hrpc", 4) == 0 ) {

		data_ += 5;	len_ -= 5;	// skip "hrpt" + 1 byte of version
		
		int head_len = 0;
		readInt(head_len);

		if( !skip(head_len) ) {		// skip "org.apache.hadoop.security.UserGroupInformation3org.apache.hadoop.security.UnixUserGroupInformation STRING_UGInorthwestern_list1users"					
			cout << "Failed to skip " << head_len << " bytes" << endl;
			return false;
		}
	}
	return true;
}


bool JStream::readBool(bool &value)
{
	if(len_ <= 0)
		return false;

	value = *data_;
	data_++; len_--;
	return true;
}

bool JStream::readBytes(char *value, unsigned int len)
{
	if(len_ < len)
		return false;
	for(int i=0; i<len; i++)
		value[i] = data_[i];
	
	data_+=len; len_-=len;
	return true;
}

bool JStream::readShort(short & value)
{
	const int vlen = 2;
	if(len_ < vlen)
		return false;

	value = ::readShort(data_);
	data_ += vlen;
	len_ -= vlen;
	return true;
}

bool JStream::readUShort(unsigned short & value)
{
	const int vlen = 2;
	if(len_ < vlen)
		return false;

	value = ::readUShort(data_);
	data_ += vlen;
	len_ -= vlen;
	return true;
}

bool JStream::readInt(int & value)
{
	const int vlen = 4;
	if(len_ < vlen)
		return false;

	value = ::readInt(data_);
	data_ += vlen;
	len_ -= vlen;
	return true;
}

bool JStream::readFloat(float & value)
{
	const int vlen = 4;
	if(len_ < vlen)
		return false;

	value = ::readFloat(data_);
	data_ += vlen;
	len_ -= vlen;
	return true;
}

bool JStream::readVInt(int & value)
{
	int slen = 0;
	value = ::readVInt(data_, slen);
	if(slen > len_)
		return false;
	data_ += slen;
	len_ -= slen;
	return true;
}

bool JStream::readLong(long long &value)
{
	const int vlen = 8;
	if(len_ < vlen)
		return false;

	value = ::readLong(data_);
	data_ += vlen;
	len_ -= vlen;
	return true;
}

bool JStream::readClassname(string &value)
{
	unsigned short nbytes = 0;
	
	if( !readUShort(nbytes) )
		return false;
	
	value.assign(data_, nbytes);
	data_ += nbytes;	len_ -= nbytes;
	return true;
}

bool JStream::readString(string &value)
{
	int nbytes = 0;
	
	if( !readVInt(nbytes) )
		return false;

	if(nbytes > 0) {
		value.assign(data_, nbytes);
		data_ += nbytes;	len_ -= nbytes;
	}
	
	return true;
}

bool JStream::readJavaLangString(string &value)
{
	string javalang;

	readClassname(javalang);
	if(javalang != "java.lang.String") {
		cerr << "readJavaLangString error: the stream should contain java.lang.String first, but now it has " << javalang << endl;
		return false;
	}

	return readClassname(value);
	return true;
}



bool ID::Read(JStream &jstream)
{
	return jstream.readInt(id_);
}

CharVect ID::getID()
{
	vector<char> vid;
	push_vect_int_reverse(vid, id_);
	return vid;
}

bool JobID::Read(JStream &jstream)
{
	if(!ID::Read(jstream))
		return false;
	
	if(! jstream.readString(jtId_) )
		return false;

	if( !jstream.readBool(isMap_) )
		return false;

	return true;
}


CharVect JobID::getJobID()
{
	vector<char> vid;
	// Java order is different
	push_vect_int_reverse(vid, id_);

	push_vect_string(vid, jtId_);

	return vid;
}

bool Counters::skip(JStream &stream)
{
	stream.readInt(numClasses_);

	if( numClasses_ != 0 ) {
		cout << " *************************** Cannot deal with counters ********************* " << endl;
		return false;
	}
	
	while(numClasses_-- > 0) {
		string groupName;
		stream.readString(groupName);

		// In Group.readFields
		string displayName;
		stream.readString(displayName);

		
	}

	return true;
}

bool TaskStatus::Read(JStream &jstream)
{
	// Read taskID (TaskAttemptID.read)
		// ID.readFields ( => Dataoutput.readInt )			
	jstream.readInt(taskAttemptID_.id_);			
		// TaskID.read
			// ID.readFields ( => Dataoutput.readInt )
	jstream.readInt(taskID_.id_);					
		// JobID.read
	jobID_.Read(jstream);

	jstream.readFloat(progress_);

	jstream.readString(runState_);
	jstream.readString(diagnosticInfo_);
	jstream.readString(stateString_);	
	jstream.readString(phase_);

	jstream.readLong(startTime_);
	jstream.readLong(finishTime_);

	jstream.readBool(includeCounters_);
	
	return true;
}

CharVect TaskStatus::getTaskID()
{
	vector<char> vid;
	// Java order is different
	push_vect_vector(vid, taskAttemptID_.getID());

	push_vect_vector(vid, taskID_.getID());

	push_vect_vector(vid, jobID_.getJobID());

	vid.push_back( jobID_.isMap_ );

	return vid;
}

bool Task::Read(JStream &jstream)
{
	// Read job file job.xml
	jstream.readString(jobFile_);
	
	// Read taskID (TaskAttemptID.read)
		// ID.readFields ( => Dataoutput.readInt )			
	jstream.readInt(taskAttemptID_.id_);			
		// TaskID.read
			// ID.readFields ( => Dataoutput.readInt )
	jstream.readInt(taskID_.id_);					
		// JobID.read
	jobID_.Read(jstream);
					
	jstream.readInt(partition_);

	jstream.readString(outPath_);

	taskStatus_.Read(jstream);

	return true;
}

bool TaskTrackerStatus::Read(JStream &stream) 
{
	stream.readClassname(trackerName_);
	stream.readClassname(host_);
			
	stream.readInt(httpPort_);
	stream.readInt(failures_);
	stream.readInt(maxMapTasks_);
	stream.readInt(maxReduceTasks_);
	
	taskReports_.clear();
	int numTasks = 0;
	stream.readInt(numTasks);

	for (int i = 0; i < numTasks; i++) {
		TaskStatus ts;
		bool bIsMap;

		stream.readBool(bIsMap);
		ts.Read(stream);
		
		taskReports_.push_back(ts);
	}

	return true;
}

bool InterTrackerProtocol_Heartbeat::Read(JStream &stream) 
{
	// skip total_len (int) and ipc_id (int)
	stream.skip(8);

	string heartbeat;
	stream.readClassname(heartbeat);
	if(heartbeat != "heartbeat")
		return false;

	int num_parameter = 0;
	stream.readInt(num_parameter);
	if(num_parameter != 4)	// Four parameters
		return false;
	
	string pname;
	stream.readClassname(pname);
	if(pname != "org.apache.hadoop.mapred.TaskTrackerStatus")
		return false;
	stream.readClassname(pname);		// skip the class name again

	status_.Read(stream);

	// Then read the last 3 parameters
	// Skip now
	// boolean initialContact, boolean acceptNewTasks, short responseId
	string para_class;
	stream.readClassname(para_class);
//printf("Param1: %s\n", para_class.c_str());	
	if(para_class != "boolean") 
		return false;
	stream.readBool(initialContact_);
	
	stream.readClassname(para_class);
//printf("Param2: %s\n", para_class.c_str());	
	if(para_class != "boolean") 
		return false;
	stream.readBool(acceptNewTasks_);

	stream.readClassname(para_class);
//printf("Param3: %s\n", para_class.c_str());		
	if(para_class != "short") 
		return false;
	stream.readShort(responseId_);
		
	return true;
}

///////////////////////////////////////////////////////////////////////////

bool Block::Read(JStream &stream)
{	
	stream.readLong(blkid_);
	stream.readLong(len_);
	// Not in version 0.14.0
	if(!bOldVersion)
		stream.readLong(generationStamp_);

	return true;
}

bool DatanodeID::Read(JStream &stream)
{
	stream.readClassname(name_);
	stream.readClassname(storageID_);
	unsigned short infoPort = 0;
	stream.readUShort(infoPort);
	infoPort_ = infoPort;
	cout << "name_:" << name_ << endl;
	return true;
}


bool DatanodeInfo::Read(JStream &stream)
{
	DatanodeID::Read(stream);

	short ipcPort;
	stream.readShort(ipcPort);
	ipcPort_ = ipcPort & 0x0000ffff;

	stream.readLong(capacity_);
	stream.readLong(dfsUsed_);
	stream.readLong(remaining_);
	stream.readLong(lastUpdate_);

	stream.readInt(xceiverCount_);
	stream.readString(location_);
	stream.readString(hostName_);
	stream.readString(adminStates_);
	cout << "hostName_: " << hostName_ << endl;
	return true;
}

bool LocatedBlock::Read(JStream &stream)
{
	// Not in version 0.14.0
	if(!bOldVersion)
		stream.readBool(corrupt_);
	stream.readLong(offset_);
	b_.Read(stream);
	int count = 0;
	stream.readInt(count);
	for (int i = 0; i < count; i++) {
		DatanodeInfo di;
		di.Read(stream);
		locs_.push_back(di);
	}
		
	return true;
}

bool LocatedBlocks::Read(JStream &stream)
{	
	stream.readLong(fileLength_);
	// Not in version 0.14.0
	if(!bOldVersion)
		stream.readBool(underConstruction_);

	int nrBlocks = 0;
	stream.readInt(nrBlocks);
	
	//cout << "fileLength_: " << fileLength_ << " nrBlocks: " << nrBlocks << endl;
	
	for (int i = 0; i < nrBlocks; i++) {		
		LocatedBlock lb;
		lb.Read(stream);
		blocks_.push_back(lb);
	}
	
	return true;
}

bool DFSFileInfo::Read(JStream &stream)
{
	
	stream.readString(path_);
	stream.readLong(length_);
	//cout << "File: " << path_ << " with lenght " << length_ << endl;

	stream.readBool(isdir_);
	stream.readShort(block_replication_);
	stream.readLong(blocksize_);
	stream.readLong(modification_time_);
	stream.readShort(permission_);
	stream.readString(owner_);
	stream.readString(group_);


	Print();
	return true;
}

void DFSFileInfo::Print()
{
	cout << "File: " << path_ << " with lenght " << length_ << endl;
	cout << "Block replication: " << block_replication_ << " and block size is " << blocksize_ << endl;
	cout << "Owner: " << owner_ << " and group: " << group_ << endl;
}

////////////////////////////////////////////////////////////////////////////

string ipc_func_name(char *payload, int plen)
{
	// skip total_len (int) and ipc_id (int)
	payload += 8;
	plen -= 8;

	int len = readShort(payload);
	payload += 2;
	plen -= 2;

	if( plen < len) {
		return string("");
	}

	string s;
	s.assign(payload, len);
	return s;
}

string ipc_func_name(JStream &stream)
{
	// skip total_len (int) and ipc_id (int)
	stream.skip(8);
	string s;
	stream.readClassname(s);

	return s;
}

CharVect ipc_call_id(char *pkt, int pkt_len)
{
	struct ip *iph = (struct ip*) pkt;
	//printf("%d %d\n", iph->ip_hl, iph->ip_p);
	//pkt += iph->ip_hl * 4;		// skip IP header. Should be changed to read the length field in IP header

	pkt = ip_payload(iph);

	if( strncmp(pkt, "hrpc", 4) == 0 ) {	// The header file
		pkt += 5;	// skip "hrpt" + 1 byte of version

		int head_len = readInt(pkt);
		pkt += (head_len + 4);					// 4 bytes for the IRC call length
		pkt_len -= (5+4+head_len);

		cout << "Head len: " << head_len << endl;
	}

	if(pkt_len >= 8) {
		// skip 4 bytes of len and return the next 4 bytes
		return nchars_to_charvect(pkt+4, 4);
	}
	else {
		cout << "Wrong format, may be fragmented" << endl;
	}		

	return CharVect();
}

/*
char *skip_ipc_header(char *payload, int &plen)
{
	char *pkt = payload;
	
	if( strncmp(pkt, "hrpc", 4) == 0 ) {	// The header file
		pkt += 5;					// skip "hrpt" + 1 byte of version

		int head_len = readInt(pkt);
		pkt += (head_len + 4);					// 4 bytes for the IRC call length
		plen -= (5+4+head_len);

		//cout << "Head len: " << head_len << endl;
	}

	return pkt;
}
*/

CharVect ipc_return_id(char *pkt, int iplen)
{
	int plen = iplen;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	cout << "plen: " << plen << endl;
	if(plen > 0) {
		return nchars_to_charvect(payload, 4);
	}

	return CharVect();
}


CharVect sock_id(char *pkt, int pkt_len)
{
	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);
	if(plen > 0)	
		id =sock_vect(iph);
	return id;	
}

vector<CharVect> sock_reverse_id(char *pkt, int pkt_len)
{
	vector<CharVect> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);
	if(plen > 0)	
		id.push_back( sock_vect_reverse(iph) );
	return id;	
}

CharVect ipc_getFileInfo(char *pkt, int iplen)
{
	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);
	
	JStream jstream(payload, plen);
	
	if(!jstream.skip_hrpc_header())
		return id;
	
	jstream.skip(8);	// skip IPC_call_length (int) and IPC_ID (int)

	string funcname;
	jstream.readClassname(funcname);
	if(funcname != "getFileInfo") {
		cerr << "Error format of message getFileInfo. Function name is not correct: " << funcname << endl;
		return id;
	}

	int num_param;	// 
	jstream.readInt(num_param);
	if( num_param != 1 ) {
		cerr << "Error format of message getFileInfo. Number of parameter is " << num_param << endl;
		return id;
	}
	
	string str;
	jstream.readClassname(str);		// should be "java.lang.String"
	jstream.readClassname(str);		// the file name
	push_vect_string(id, str);

	return id;
}

CharVect ipc_getBlockLocations(char *pkt, int iplen)
{
	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);
	
	JStream jstream(payload, plen);
	
	if(!jstream.skip_hrpc_header())
		return id;
	
	jstream.skip(8);	// skip IPC_call_length (int) and IPC_ID (int)

	string funcname;
	jstream.readClassname(funcname);
	if(!bOldVersion) {
		if(funcname != "getBlockLocations") {
			cerr << "Error format of message getBlockLocations. Function name is not correct: " << funcname << endl;
			return id;
		}
	}
	else {
		if(funcname != "open") {
			cerr << "Error format of message open. Function name is not correct: " << funcname << endl;
			return id;
		}
	}

	int num_param;	// 
	jstream.readInt(num_param);
	if( num_param != 3 ) {
		cerr << "Error format of message getBlockLocations. Number of parameter is " << num_param << endl;
		return id;
	}
	
	string str;
	jstream.readClassname(str);		// should be "java.lang.String"
	jstream.readClassname(str);		// the file name
	push_vect_string(id, str);

	return id;
}

CharVect DFSFileInfo_ID(char *pkt, int iplen)
{
	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);
	
	JStream jstream(payload, plen);
	
	jstream.skip(4);	// skip IPC_ID
	bool bIPCCall = false;

	jstream.readBool(bIPCCall);
	if(bIPCCall != 0)	// IPC call error
		return id;

	string objname;
	jstream.readClassname(objname);
	if(objname != "org.apache.hadoop.dfs.DFSFileInfo") {
		cerr << "Error format of object DFSFileInfo. Object name is not correct: " << objname << endl;
		return id;
	}
	jstream.readClassname(objname);	// the objname again

	DFSFileInfo dfi;
	if( dfi.Read(jstream) ) {
		id = string_to_charvect(dfi.path_);
	}
	
	return id;
}

vector<CharVect> IPC_LocatedBlocks(char *pkt, int iplen)
{
	vector<CharVect> cvs;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	JStream jstream(payload, plen);

	jstream.skip(4);	// skip IPC_ID
	bool bIPCCall = false;

	jstream.readBool(bIPCCall);
	if(bIPCCall != 0)	// IPC call error
		return cvs;

	string objname;
	jstream.readClassname(objname);
	if(objname != "org.apache.hadoop.dfs.LocatedBlocks")
		return cvs;

	jstream.readClassname(objname);	// skip "org.apache.hadoop.dfs.LocatedBlocks" again

	LocatedBlocks lbs;
	lbs.Read(jstream);

	for(int i=0; i<lbs.blocks_.size(); i++) {
		char *s = (char *) &lbs.blocks_[i].b_.blkid_;
		cvs.push_back(  nchars_to_charvect(s, 8) );
	}

	return cvs;
}

CharVect DFSClient_BlockReader(char *pkt, int iplen)
{

	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	if(!bOldVersion) {
		if(plen != 35)
			return id;		
	}
	else {
		if(plen != 27)
			return id;		
	}
	
	JStream jstream(payload, plen);
	short version;
	jstream.readShort(version);

	char op;
	jstream.readBytes(&op, 1);
	if(op != 0x51)
		return id;

	long long bid;
	jstream.readLong(bid);
	char *s = (char *) &bid;
	id =  nchars_to_charvect(s, 8);
	
	return id;
}

bool Is_ReadBlock(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	if(!bOldVersion) {
		if(plen != 35)
			return false;		
	}
	else {
		if(plen != 27)
			return false;		
	}
	return true;
}

CharVect DFSClient_BlockReader_Done(char *pkt, int iplen)
{
	vector<char> id;
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	if(plen == 4) {
		JStream jstream(payload, plen);
		int n = 0;
		if(jstream.readInt(n)) {
			if(n == 0) {
				for(int i=0; i<4; i++)
					id.push_back('0');
			}
		}
	}
	return id;	
}

string read_ipc_func_name(char *pkt, int iplen, JStream &stream)
{
	string s = "";
	int plen = 0;
	char *payload = ip_payload((struct ip*) pkt, iplen, plen);

	stream = JStream(payload, plen);
	
	if(!stream.skip_hrpc_header())
		return s;

	// skip total_len (int) and ipc_id (int)
	stream.skip(8);
	stream.readClassname(s);
	return s;
}

CharVect get_filename_from_put_ipc_calls(char *pkt, int iplen, char *target_func_name)
{
	vector<char> id;
	string funcname;
	JStream stream;
	
	funcname = read_ipc_func_name(pkt, iplen, stream);
	if(funcname != target_func_name)
		return id;

	int num_param = 0;
	stream.readInt(num_param);
	if(num_param <= 0) {
		cerr << "In get_filename_from_put_ipc_calls: ipc call has " << num_param << " parameters" << endl;
		return id;
	}
	
	string filename;
	if( stream.readJavaLangString(filename) ) {
		id = string_to_charvect(filename);
	}

	return id;
}

CharVect ipc_create(char *pkt, int iplen)
{
	return get_filename_from_put_ipc_calls(pkt, iplen, "create");
}

CharVect ipc_addblock(char *pkt, int iplen)
{
	return get_filename_from_put_ipc_calls(pkt, iplen, "addBlock");		
}

CharVect ipc_complete(char *pkt, int iplen)
{
	return get_filename_from_put_ipc_calls(pkt, iplen, "complete");		
}

CharVect ipc_setreplication(char *pkt, int iplen)
{
	return get_filename_from_put_ipc_calls(pkt, iplen, "setReplication");		
}

CharVect ipc_setpermission(char *pkt, int iplen)
{
	return get_filename_from_put_ipc_calls(pkt, iplen, "setPermission");		
}


////////////////////////////////////////////////////////////////////////////

CharVect read_JobID(char *stream, int plen)
{
	char *jobclassname = "org.apache.hadoop.mapred.JobID";

	// skip the jobclassname
	int len = readShort(stream), slen = strlen(jobclassname);
	stream+=2+len;
	plen -= 2+len;

	// skip the string len (short)
	len = readShort(stream), slen = strlen(jobclassname);
	stream+= 2;
	plen -= 2;

	for(int i=0; i<10; i++)
		printf("%02x ", stream[i]);
	printf("\n");

	//cout << "len: " << len << ", slen: " << slen << endl;
	if(len == slen && strncmp(stream, jobclassname, len) == 0 ) {
		stream+=len;	// skip the jobclassname again
		plen -= len;	

		vector<char> id = nchars_to_charvect(stream, 4);	// Int of Job ID
		stream += 4;
		plen -= 4;

		int vlen = 0;
		int jlen = readVInt(stream, vlen);
		stream += vlen;
		plen -= vlen;

		if(jlen > plen) {
			cout << "Potential error in func new_job_id: left payload length is " << plen << " but the common job name length is " << jlen << endl;
		}

		for(int i=0; i<jlen; i++)
			id.push_back( stream[i] );

		return id;
	}

	return CharVect();
}

vector<CharVect> new_job_id(char *pkt, int iplen)
{
	vector<CharVect> follow_ids;
	int plen = iplen;
	char *jobclassname = "org.apache.hadoop.mapred.JobID";
	
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	//pkt = skip_ipc_header(payload, plen);

	payload += 4;		// skip IPC_ID
	plen -= 4;
	if( *payload != 0 ) {	// IPC call error
		return follow_ids;
	}
	else {
		payload++;	// skip the call return flag
		plen--;

	/*
		// skip the jobclassname
		int len = readShort(payload), slen = strlen(jobclassname);
		payload+=2+len;
		plen -= 2+len;

		// skip the string len (short)
		len = readShort(payload), slen = strlen(jobclassname);
		payload+= 2;
		plen -= 2;

		//cout << "len: " << len << ", slen: " << slen << endl;
		if(len == slen && strncmp(payload, jobclassname, len) == 0 ) {
			payload+=len;	// skip the jobclassname again
			plen -= len;	

			vector<char> id = nchars_to_charvect(payload, 4);	// Int of Job ID
			payload += 4;
			plen -= 4;

			int vlen = 0;
			int jlen = readVInt(payload, vlen);
			payload += vlen;
			plen -= vlen;

			if(jlen != plen) {
				cout << "Potential error in func new_job_id: left payload length is " << plen << " but the common job name length is " << jlen << endl;
			}

			for(int i=0; i<jlen; i++)
				id.push_back( payload[i] );
		*/

			CharVect id = read_JobID(payload, plen);
			if(id.size() > 0) {
				follow_ids.push_back( id );

				// Add Follow_ID for  "getSystemDir" IPC call
				id.clear();
				id = sock_vect(iph);
				push_vect_chars(id, "getSystemDir");
				follow_ids.push_back( id );
			}
		
	}
	
	return follow_ids;
}


CharVect ipc_getSystemDir(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id = sock_vect_reverse(iph);
	char * targetname = "getSystemDir";
	string funcname = ipc_func_name(payload, plen);

	if( funcname == targetname) {
		push_vect_chars(id, targetname);	
	}
	else
		id.clear();

	return id;
}

char* hadoop_skip_text(char *stream, int &len)
{
	int txtlen = readShort(stream);

	len -= 2 + txtlen;
	return (stream+2+txtlen);
}

CharVect get_jobid_from_mkdirs_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "mkdirs";
	string funcname = ipc_func_name(payload, plen);

	cout << funcname << endl;
	
	if( funcname == targetname) {
		// skip total_len and ipc_id
		payload += 8;
		plen -= 8;

		// skip the function name
		payload = hadoop_skip_text(payload, plen);

		// skip number of param (int) of the function
		payload += 4;
		plen -= 4;

		// skip java.lang.string
		payload = hadoop_skip_text(payload, plen);

		// the dir name
		int nlen = readShort(payload);
		payload += 2;
		plen -= 2;

		string dirname;
		dirname.assign( payload, nlen );
		size_t pos = dirname.find_last_of('/');
		if( pos != string::npos )
			dirname = dirname.substr( pos+1 );

		cout << dirname << endl;
		if(dirname.substr(0, 4) == "job_") {
			dirname = dirname.substr(4);

			cout << dirname << endl;
		
			pos = dirname.find_first_of('_');

			if(pos != string::npos) {
				string jobname = dirname.substr(0, pos);
				string jobid = dirname.substr(pos+1);
				int nid = atoi(jobid.c_str());
				push_vect_int(id, htonl(nid));
				push_vect_string(id, jobname);

				return id;
			}
		}
	}

	return id;
}

CharVect get_jobid_from_create_addblock_complete_ipc(char *pkt, int iplen, char* funcname, char*  filename)
{
	int plen = 0;
	char *payload = ip_payload((struct ip*) pkt, iplen, plen);

	vector<char> id;
	string fname = ipc_func_name(payload, plen);

	if( funcname == funcname) {
		return get_jobid_from_upload_file_ipc(payload, plen, filename);		
	}

	return id;
}

CharVect get_jobid_from_upload_file_ipc(char *payload, int plen, const char *fname)
{
	CharVect id;

	// skip total_len and ipc_id
	payload += 8;
	plen -= 8;

	// skip the function name
	payload = hadoop_skip_text(payload, plen);

	// skip number of param (int) of the function
	payload += 4;
	plen -= 4;

	// skip java.lang.string
	payload = hadoop_skip_text(payload, plen);

	// the dir name
	int nlen = readShort(payload);
	payload += 2;
	plen -= 2;

	string dirname, filename;
	dirname.assign( payload, nlen );
	size_t pos = dirname.find_last_of('/');
	if( pos != string::npos ) {
		filename = dirname.substr( pos+1 );
		dirname = dirname.substr(0, pos);
	}
	else
		return CharVect();

	//if(filename != "job.jar")
	if(filename != fname)
		return CharVect();

	pos = dirname.find_last_of('/');
	if( pos != string::npos )
		dirname = dirname.substr( pos+1 );
	
	cout << dirname << endl;
	if(dirname.substr(0, 4) == "job_") {
		dirname = dirname.substr(4);

		cout << dirname << endl;
	
		pos = dirname.find_first_of('_');

		if(pos != string::npos) {
			string jobname = dirname.substr(0, pos);
			string jobid = dirname.substr(pos+1);
			int nid = atoi(jobid.c_str());
			push_vect_int(id, htonl(nid));
			push_vect_string(id, jobname);

			return id;
		}
	}

	return CharVect();
}

CharVect get_jobid_from_create_job_jar_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "create";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.jar");
		/*
		// skip total_len and ipc_id
		payload += 8;
		plen -= 8;

		// skip the function name
		payload = hadoop_skip_text(payload, plen);

		// skip number of param (int) of the function
		payload += 4;
		plen -= 4;

		// skip java.lang.string
		payload = hadoop_skip_text(payload, plen);

		// the dir name
		int nlen = readShort(payload);
		payload += 2;
		plen -= 2;

		string dirname, filename;
		dirname.assign( payload, nlen );
		size_t pos = dirname.find_last_of('/');
		if( pos != string::npos ) {
			filename = dirname.substr( pos+1 );
			dirname = dirname.substr(0, pos);
		}
		else
			return CharVect();

		if(filename != "job.jar")
			return CharVect();

		pos = dirname.find_last_of('/');
		if( pos != string::npos )
			dirname = dirname.substr( pos+1 );
		
		cout << dirname << endl;
		if(dirname.substr(0, 4) == "job_") {
			dirname = dirname.substr(4);

			cout << dirname << endl;
		
			pos = dirname.find_first_of('_');

			if(pos != string::npos) {
				string jobname = dirname.substr(0, pos);
				string jobid = dirname.substr(pos+1);
				int nid = atoi(jobid.c_str());
				push_vect_int(id, htonl(nid));
				push_vect_string(id, jobname);

				return id;
			}
		}
		*/
	}

	return id;
}



CharVect get_jobid_from_addblock_job_jar_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "addBlock";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.jar");		
	}

	return id;
}

CharVect get_jobid_from_complete_job_jar_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "complete";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.jar");		
	}

	return id;
}

CharVect get_jobid_from_setreplication_job_jar_ipc(char *pkt, int iplen)
{
	return get_jobid_from_create_addblock_complete_ipc(pkt, iplen, "setReplication", "job.jar");
}

CharVect get_jobid_from_setpermission_job_jar_ipc(char *pkt, int iplen)
{
	return get_jobid_from_create_addblock_complete_ipc(pkt, iplen, "setPermission", "job.jar");
}

CharVect get_jobid_from_create_job_xml_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "create";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.xml");		
	}

	return id;
}

CharVect get_jobid_from_addblock_job_xml_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "addBlock";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.xml");		
	}

	return id;
}

CharVect get_jobid_from_complete_job_xml_ipc(char *pkt, int iplen)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, iplen, plen);

	vector<char> id;
	char * targetname = "complete";
	string funcname = ipc_func_name(payload, plen);

	//cout << funcname << endl;
	
	if( funcname == targetname) {
		return get_jobid_from_upload_file_ipc(payload, plen, "job.xml");		
	}

	return id;
}

CharVect get_jobid_from_create_job_split_ipc(char *pkt, int iplen)
{	
	return get_jobid_from_create_addblock_complete_ipc(pkt, iplen, "create", "job.split");
}

CharVect get_jobid_from_addblock_job_split_ipc(char *pkt, int iplen)
{
	return get_jobid_from_create_addblock_complete_ipc(pkt, iplen, "addBlock", "job.split");
}

CharVect get_jobid_from_complete_job_split_ipc(char *pkt, int iplen)
{
	return get_jobid_from_create_addblock_complete_ipc(pkt, iplen, "complete", "job.split");
}




CharVect submitJob_ipc(char *pkt, int pkt_len)
{
	struct ip *iph = (struct ip*) pkt;
	pkt = ip_payload(iph);

	if( strncmp(pkt, "hrpc", 4) == 0 ) {	// The header file
		pkt += 5;	// skip "hrpt" + 1 byte of version

		int head_len = readInt(pkt);
		pkt += (head_len + 4);					// 4 bytes for the IRC call length
		pkt_len -= (5+4+head_len);

		//cout << "Head len: " << head_len << endl;
	}

	char *payload = pkt;
	int plen = pkt_len;

	char * targetname = "submitJob";
	string funcname = ipc_func_name(payload, plen);
	
	if( funcname == targetname) {		
		payload += 8;		// skip total_len and IPC_ID
		plen -= 8;
		
		// skip submitJob
		payload = hadoop_skip_text(payload, plen);

		// skip number of param (int) of the function
		payload += 4;
		plen -= 4;

		CharVect id = read_JobID(payload, plen);
		return id;
	}

	return CharVect();
}


/////// HeartBeatResponse //////////////

CharVect HeartBeatResponse_Launch_Task(char *pkt, int pkt_len)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);

	JStream jstream(payload, plen);

	jstream.skip(4);	// skip IPC_ID
	bool bIPCCall = false;

	jstream.readBool(bIPCCall);
	if(bIPCCall != 0)	// IPC call error
		return CharVect();
	
	string objname;
	jstream.readClassname(objname);
	if(objname != "org.apache.hadoop.mapred.HeartbeatResponse")
		return CharVect();

	jstream.readClassname(objname);	// skip "org.apache.hadoop.mapred.HeartbeatResponse" again

	// See HeartBeatResponse.write
	// skip responseId (short) and heartbeatInterval (int);
	//jstream.skip(6);
	short responseId = 0;
	int heartbeatInterval;
	jstream.readShort(responseId);
	jstream.readInt(heartbeatInterval);

	int nactions = 0;
//cout << "nactions: " << nactions << endl;
	jstream.readVInt(nactions);

	if(nactions == 0) {
		CharVect sockid =  sock_vect_reverse(iph);
		char text[256];
		sprintf(text, "heatbeat_response_id_%04x\n", responseId-1);
//printf("In HeartBeatResponse_Launch_Task, responseId: %d\n", responseId);
		push_vect_chars(sockid, text);
cout << "HeartBeatResponse_Launch_Task: " << charvect_to_string(sockid) << endl;
		return sockid;
	}
	
	for(int i=0; i<nactions; i++) {
		string action;		
		jstream.readString(action);
//cout << "action: " << action << endl;
		
		if(action == "LAUNCH_TASK") {		// See LaunchTaskAction.write
			bool bIsMap = false;
			jstream.readBool(bIsMap);
			
			// See Task.write
			Task task;
			if( task.Read(jstream) ) {
				return task.jobID_.getJobID();
			}	
		}
		else if(action == "KILL_JOB") {
			JobID jobId;
			jobId.Read(jstream);
			return jobId.getJobID();
		}
	}

	return CharVect();
}

vector<CharVect> HeartBeatResponse_Launch_Task_TaskID(char *pkt, int pkt_len)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);
	vector<CharVect> cvs;
	
	JStream jstream(payload, plen);

	jstream.skip(4);	// skip IPC_ID
	bool bIPCCall = false;

	jstream.readBool(bIPCCall);
	if(bIPCCall != 0)	// IPC call error
		return cvs;
	
	string objname;
	jstream.readClassname(objname);
	if(objname != "org.apache.hadoop.mapred.HeartbeatResponse")
		return cvs;

	jstream.readClassname(objname);	// skip "org.apache.hadoop.mapred.HeartbeatResponse" again

	// See HeartBeatResponse.write
	// skip responseId (short) and heartbeatInterval (int);
	jstream.skip(6);

	int nactions = 0;
//cout << "nactions: " << nactions << endl;
	jstream.readVInt(nactions);
	
	for(int i=0; i<nactions; i++) {
		string action;		
		jstream.readString(action);
//cout << "action: " << action << endl;
		
		if(action == "LAUNCH_TASK") {		// See LaunchTaskAction.write
			bool bIsMap = false;
			jstream.readBool(bIsMap);
			
			// See Task.write
			Task task;
			if( task.Read(jstream) ) {
				//return task.taskStatus_.getTaskID();
				cvs.push_back(  task.taskStatus_.getTaskID() );
				return cvs;
			}
			
			
		}
	}

	return cvs;
}

CharVect heartbeat_status_succeeded(char *pkt, int pkt_len)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);

	JStream jstream(payload, plen);
	jstream.skip_hrpc_header();
	
	InterTrackerProtocol_Heartbeat hb;
	if( !hb.Read(jstream) )
		return CharVect();

	for(int i = 0; i< hb.status_.taskReports_.size(); i++) {
		TaskStatus &ts = hb.status_.taskReports_[i];

		//return ts.jobID_.getJobID();
		return ts.getTaskID();
	}
}


CharVect heartbeat_status_running(char *pkt, int pkt_len)
{
	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);

	JStream jstream(payload, plen);
	jstream.skip_hrpc_header();
	
	InterTrackerProtocol_Heartbeat hb;
	if( !hb.Read(jstream) )
		return CharVect();

//printf("In heartbeat_status_running, response id %04x\n", hb.responseId_);

	for(int i = 0; i< hb.status_.taskReports_.size(); i++) {
		TaskStatus &ts = hb.status_.taskReports_[i];

		//return ts.jobID_.getJobID();
		return ts.getTaskID();
	}
}

vector<CharVect> heartbeat_status_running_follow_ids(char *pkt, int pkt_len)
{
	vector<CharVect> cvs;
//printf("In heartbeat_status_running_follow_ids\n");

	int plen = 0;
	struct ip *iph = (struct ip*) pkt;
	char *payload = ip_payload(iph, pkt_len, plen);

	JStream jstream(payload, plen);
	jstream.skip_hrpc_header();
	
	InterTrackerProtocol_Heartbeat hb;
	if( !hb.Read(jstream) )
		return cvs;

	CharVect sockid =  sock_vect(iph);
	char text[256];
	sprintf(text, "heatbeat_response_id_%04x\n", hb.responseId_);
//printf("In heartbeat_status_running_follow_ids, text: %s\n", text);
	//cvs.push_back( chars_to_charvect(text) );
	push_vect_chars(sockid, text);
	cvs.push_back(sockid);
cout << "heartbeat_status_running_follow_ids: " << charvect_to_string(sockid) << endl;
	
	for(int i = 0; i< hb.status_.taskReports_.size(); i++) {
		TaskStatus &ts = hb.status_.taskReports_[i];

		//return ts.jobID_.getJobID();
		cvs.push_back( ts.getTaskID() );
	}

	return cvs;
}

/*
vector<CharVect> Follow_IDs(char *pkt, int pkt_len)
{
	struct dns_message dmsg;
	vector<CharVect> follow_ids;
	
	dns_decode_message(&dmsg, &pkt);

	for(int i=0; i<dmsg.header.ancount; i++) {
		follow_ids.push_back( chars_to_charvect( dmsg.answer[i].name) );
	}

	return follow_ids;
}
*/

