/*
 * Tcpdump packet handler
 * Author: 
 *	Yao Zhao
 * Goal:
 *	Handle tcpdump files
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "handle_pcap.h"
#include "util.h"

namespace __gnu_cxx                                                                                 
{                                                                                             
	template<> struct hash< std::string >                                                       
	{                                                                                           
		size_t operator()( const std::string& x ) const                                           
		{                                                                                         
			return hash< const char* >()( x.c_str() );                                              
		}                                                                                         
	};

}

HandlePcap::HandlePcap():pcap_(NULL),status_(0),hdr_size_(-2)
{
	avg_delay_ = 0;
	lamda_ = 0;
	delay_sample_num_ = 10000;

	local_address_ = INADDR_BROADCAST;
	
	pcap_fname_ = NULL;
	pcap_filter_ = NULL;

	spontaneous_threshold_ = 2;
}

HandlePcap::~HandlePcap()
{
	if(pcap_)
		pcap_close(pcap_);
}

int HandlePcap::pcap_hdr_size()
{
	if(hdr_size_ == -2)
		if(status_ > 0)
			hdr_size_ = pcap_hdr_size(pcap_);

	return hdr_size_;
}

int HandlePcap::pcap_hdr_size(pcap_t* pcap)
{
	int dl = pcap_datalink(pcap);
	int hdr_size = -1;

	switch (dl) {
	case DLT_NULL:
		hdr_size = 4;
		break;

	case DLT_EN10MB:
		hdr_size = 14;
		break;

	case DLT_FDDI:
		hdr_size = 13 + 8;	// fddi_header + LLC
		break;

	case DLT_RAW:
		hdr_size = 0;
		break;

	default:
		fprintf(stderr, "unknown data link type 0x%x", dl);
		break;
	}
	return hdr_size; 
}

bool HandlePcap::open_pcap_file(char* &pkt_filename, char* &pcap_filter)
{	
	struct bpf_program bpf_prog;
	//struct pcap_pkthdr hdr;
	//const u_char *pkt;
	int hdr_size;
	
	if (!(pcap_ = pcap_open_offline(pkt_filename, pcap_errbuf_))) {
		fprintf(stderr, "can't open file %s\n", pkt_filename);
		pcap_perror(pcap_, "pcap_open_offline");
		return false;
	}

	if ((hdr_size = pcap_hdr_size(pcap_)) < 0) {
		pcap_close(pcap_);
		return false;
	}

	if (pcap_filter) {
		if (pcap_compile(pcap_, &bpf_prog, pcap_filter, 1, 0xffff0000)) {
			pcap_perror(pcap_, "pcap_compile");
			return false;
		}
		else if (pcap_setfilter(pcap_, &bpf_prog)) {
			pcap_perror(pcap_, "pcap_setfilter");
			return false;
		}
	}

	if( pcap_hdr_size() < 0 )
		status_ = -1;

	status_ = 1;

	return true;
}

bool HandlePcap::close_pcap_file()
{
	if(pcap_)
		pcap_close(pcap_);
	pcap_ = NULL;
	status_ = 0;
	hdr_size_ = -2;

	return true;
}

const u_char* HandlePcap::next_pkt(struct pcap_pkthdr *hdr)
{
	const u_char *pkt;
	pkt = pcap_next(pcap_, hdr);
	return pkt;
}

in_addr_t HandlePcap::determine_local_address()
{
	int i = 0;
	const u_char *pkt;
	struct pcap_pkthdr hdr;
	//struct miphdr *iph;
	struct ip *iph;
	int hdr_size = 0;

	hash_map<in_addr_t, int> hmap;
	hash_map<in_addr_t, int>::iterator it;

	if(local_address_ != INADDR_BROADCAST)
		return true;

	if(status_ > 0)
		close_pcap_file();

	if(! open_pcap_file(pcap_fname_, pcap_filter_) ) 
		return INADDR_BROADCAST;

	hdr_size = pcap_hdr_size();
	if(hdr_size < 0)
		return INADDR_BROADCAST;

	while(i<1000) {
		pkt = next_pkt(&hdr);

		if(!pkt)
			break;

		iph = (struct ip *) (pkt + hdr_size);

		if( (it = hmap.find( iph->ip_src.s_addr )) != hmap.end() ) {
			it->second ++;
		}
		else			
			hmap[ iph->ip_src.s_addr] = 1;

		if( (it = hmap.find( iph->ip_dst.s_addr)) != hmap.end() ) {
			it->second ++;
		}
		else			
			hmap[ iph->ip_dst.s_addr] = 1;
		i++;
	}

	int max = 0, max2;
	in_addr_t mip = INADDR_BROADCAST, mip2 = INADDR_BROADCAST;
	for(it=hmap.begin(); it != hmap.end(); it++) {
		if( it->second > max ) {
			max2 = max;
			mip2 = mip;
			
			max = it->second;
			mip = it->first;
		}
		else if(it->second > max2) {
			mip2 = it->first;
			max2 = it->second;
		}		
	}

	if(i == max && max > max2) {
		local_address_ = mip;
	}
	else {
		fprintf(stderr, "Can't determine the local IP of the pcap file. Read %d packets, and %d of them use address %s, ", i, max, inet_ntoa( *(struct in_addr *)&mip ));
		fprintf(stderr, "and %d of them use address %s, ", max2, inet_ntoa( *(struct in_addr *)&mip2 ));
	}

	close_pcap_file();

	return local_address_;
}

double HandlePcap::avg_casual_delay()
{
	int index = 0;
	const u_char *pkt;
	struct pcap_pkthdr hdr;
	//struct miphdr *iph;
	struct ip *iph;
	int hdr_size = 0;
	struct timeval now, last; //tv;
	//list<struct timeval> tlist;
	double total_delay = 0;
	int num_out_pkt = 0;

	// Make sure the file is newly opened
	if( status_ > 0 )
		close_pcap_file();

	if(! open_pcap_file(pcap_fname_, pcap_filter_) ) 
		return 0;

	hdr_size = pcap_hdr_size();
	if(hdr_size < 0)
		return 0;

	if( determine_local_address() == INADDR_BROADCAST )
		return 0;	

	last.tv_sec = -1;
	while(delay_sample_num_ >= index) {
		pkt = next_pkt(&hdr);
		if(!pkt)
			break;
		index ++;

		now = hdr.ts;
		iph = (struct ip *) (pkt + hdr_size);
		if(iph->ip_src.s_addr == local_address_) {
			if(last.tv_sec < 0)
				continue;

			double t = time_minus(now, last);
			if(t > spontaneous_threshold_)
				continue;

			total_delay += t;
			//cout << time_minus(now, last) << endl;
			num_out_pkt ++;	
		}
		else if(iph->ip_dst.s_addr == local_address_) {
			last = now;			
		}
		

		/*
		while(true) {
			//tmsg = ml.front();
			tv = tlist.front();
			
			if(  time_minus(now,  tv) > spontaneous_threshold_) {
				tlist.pop_front();
			}			
		}		
		tlist.push_back(now);
		*/
		
		
	}

	if( num_out_pkt> 0) {
		avg_delay_ = total_delay / num_out_pkt;
		lamda_ = 1/avg_delay_;
	}

	close_pcap_file();
	return avg_delay_;
}


