#ifndef __DNS_HEADER__
#define __DNS_HEADER__

#include <sys/types.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
 
 
#define PORT 53 
#define NAME_SIZE 255 
#define MAX_PACKET_SIZE 512 
#define BUF_SIZE NAME_SIZE 
#define MAX_CONCURRENT_REQUESTS 100 
 
typedef u_int16_t uint16; 
typedef u_int32_t uint32; 
 
#define NUM_RRS 5 

/*****************************************************************************/ 
union header_flags { 
  uint16 flags; 
 
  struct { 
#if BYTE_ORDER == LITTLE_ENDIAN 
    unsigned short int rcode:4; 
    unsigned short int unused:3; 
    unsigned short int recursion_avail:1; 
    unsigned short int want_recursion:1; 
    unsigned short int truncated:1; 
    unsigned short int authorative:1; 
    unsigned short int opcode:4; 
    unsigned short int answer:1; 
#else 
    unsigned short int answer:1; 
    unsigned short int opcode:4; 
    unsigned short int authorative:1; 
    unsigned short int truncated:1; 
    unsigned short int want_recursion:1; 
    unsigned short int recursion_avail:1; 
    unsigned short int unused:3; 
    unsigned short int rcode:4; 
#endif 
  } f; 
}; 

/*****************************************************************************/ 
struct dns_rr{ 
  char name[NAME_SIZE]; 
  uint16 type; 
  uint16 clas; 
  uint32 ttl; 
  uint16 rdatalen; 
  char data[NAME_SIZE]; 
}; 

/*****************************************************************************/ 
	struct dns_header_s{ 
	uint16 id; 
	union header_flags flags; 
	uint16 qdcount; 
	uint16 ancount; 
	uint16 nscount; 
	uint16 arcount; 
}; 

/*****************************************************************************/ 
struct dns_message { 
  struct dns_header_s header; 
  struct dns_rr question[NUM_RRS]; 
  struct dns_rr answer[NUM_RRS]; 
}; 


/*****************************************************************************/ 
typedef struct dns_request_s{ 
    char cname[NAME_SIZE]; 
    char ip[20]; 
    int ttl; 
    int time_pending; /* request age in seconds */ 
    int last_sent;   /* time since we last send the request to the sevrer */ 
    int server; /* index into the server list this pkt is to be sent to */ 
 
    /* the actual dns request that was recieved */ 
    struct dns_message message; 
 
    /* where the request came from */ 
    struct in_addr src_addr; 
    int src_port; 
 
    /* the orginal packet */ 
    char original_buf[MAX_PACKET_SIZE]; 
    int numread; 
    char *here; 
 
    /* next node in list */ 
    struct dns_request_s *next; 
}dns_request_t; 


/*****************************************************************************/ 
/* TYPE values */ 
enum{ A = 1,      /* a host address */ 
    NS,       /* an authoritative name server */ 
    MD,       /* a mail destination (Obsolete - use MX) */ 
    MF,       /* */ 
    CNAME,    /* the canonical name for an alias */ 
    SOA,      /* marks the start of a zone of authority  */ 
    MB,       /* a mailbox domain name (EXPERIMENTAL) */ 
    MG,       /* */ 
    MR,       /* */ 
    NUL,      /* */ 
    WKS,      /* a well known service description */ 
    PTR,      /* a domain name pointer */ 
    HINFO,    /* host information */ 
    MINFO,    /* mailbox or mail list information */ 
    MX,       /* mail exchange */ 
    TXT,      /* text strings */ 
 
    AAA = 0x1c /* IPv6 A */ 
    }; 
 
/* CLASS values */ 
enum{ 
  IN = 1,         /* the Internet */ 
    CS, 
    CH, 
    HS 
}; 
 
/* OPCODE values */ 
enum{ 
  QUERY, 
    IQUERY, 
    STATUS 
}; 
 
/* RCODE values */ 
enum{ 
    ERR_OK = 0, 
        ERR_FORMAT, 
        ERR_SRVFAIL, 
        ERR_NOEXIST, 
        ERR_NOTIMPL, 
        ERR_REFUSED 
}; 
 
#ifndef DNS_TIMEOUT  
#define DNS_TIMEOUT 60 
#endif 
 
#ifndef DNS_SEND_TIMER 
#define DNS_SEND_TIMER 2 
#endif 

/*
 * Queries are encoded such that there is and integer specifying how long 
 * the next block of text will be before the actuall text. For eaxmple:
 *             www.linux.com => \03www\05linux\03com\0
 * This function assumes that buf points to an encoded name.
 * The name is decoded into name. Name should be at least 255 bytes long.
 * RETURNS: The length of the string including the '\0' terminator.
 */
void dns_decode_name(char *name, char **buf);
/*
 * Decodes the raw packet in buf to create a rr. Assumes buf points to the 
 * start of a rr. 
 * Note: Question rrs dont have rdatalen or rdata. Set is_question when
 * decoding question rrs, else clear is_question
 * RETURNS: the amount that buf should be incremented
 */
void dns_decode_rr(struct dns_rr *rr, char **buf,int is_question,char *header);

/*
 * A raw packet pointed to by buf is decoded in the assumed to be alloced 
 * dns_message structure.
 * RETURNS: 0
 */
int dns_decode_message(struct dns_message *m, char **buf);

/* 
 * 
 */
void dns_decode_request(dns_request_t *m);


#endif

