#include <iostream>
#include <string>
#include <map>
#include "rxx.h"
#include "chdr.h"
#include "coralweb_const.h"
#include "webcache.h"
#include "httpmsg.h"
#include "httpconst.h"
#include "kademlia_coral_utils.h"
#include "coral_prot.h"


static const char *cdn_prfx      = "http";
static const char *opt_http      = "(http://)?";

static const char *cdn_str       = ".nyud.net";
static const char *cdn_url       = "\\.nyud\\.net";
static const char *cdn_url2      = "\\.http(\\.l[0-9])+\\.nyucd\\.net";
//static const char *cdn_url2    = ".http.l2.l1.l0.nyucd.net";
static       rxx   cdn_urls      ("^(.*)\\.nyu(c)?d\\.net(.*)?$");

static const char *host          = "([a-zA-Z0-9.\\-]+)";
static const char *opt_port      = "(:([0-9]+))?";

static const char *trailer       = "/(.*)";
static const char *opt_trailer   = "(/(.*))?";

//static const char *redir_host  = "localhost";
static const char *redir_prfx    = "redirect";
static const char *redir_get     = "\\?url=";
static const char *opt_redir_get = "(\\?url=)?";


//    http://www.x.com[.80].http.nyud.net:8090/index.html
static const str scoralurl_full (strbuf ("^%s%s%s%s%s$", opt_http, 
				         host, cdn_url, opt_port, trailer));
static       rxx  coralurl_full (scoralurl_full.cstr ());

static const str scoralurl2_full (strbuf ("^%s%s%s%s%s$", opt_http, 
					  host, cdn_url2, opt_port, trailer));
static       rxx  coralurl2_full (scoralurl2_full.cstr ());


//    www.x.com[.80].http.nyud.net:8090
static const str shostname_full (strbuf ("^%s%s%s$", 
					 host, cdn_url, opt_port));
static       rxx  hostname_full (shostname_full.cstr ());

//    www.x.com[.80].http.l2.l1.l0.nyucd.net:8090
static const str shostname2_full (strbuf ("^%s%s%s$", 
					  host, cdn_url2, opt_port));
static       rxx  hostname2_full (shostname2_full.cstr ());

//    http://http.nyud.net:8090/www.x.com[:80]/index.html
static const str scoralurl_split (strbuf ("^%s%s%s%s$", opt_http, 
					  host, opt_port, trailer));
static       rxx  coralurl_split (scoralurl_split.cstr ());


//    www.x.com[:80]
static const str shostname (strbuf ("^%s%s$", host, opt_port));
static       rxx  hostname (shostname.cstr ());


//    www.x.com[:80]/index.html
static const str shomeurl  (strbuf ("^%s%s%s$", host, opt_port, opt_trailer));
static       rxx  homeurl  (shomeurl.cstr ());


//    http://redirect.nyud.net:8090/?url=<URL>
static const str redirect_str   (strbuf ("%s%s", redir_prfx, cdn_str));

static const str sredirect_host (strbuf ("^%s%s%s$", 
					 redir_prfx, cdn_url, opt_port));
static       rxx  redirect_host (sredirect_host.cstr ());

static const str sredirect_sfx  (strbuf ("^/%s(.*)$", redir_get));
static       rxx  redirect_sfx  (sredirect_sfx.cstr ());

static const str sredirect_url  (strbuf ("^http://%s%s%s/%s(.*)$", redir_prfx,
					 cdn_url, opt_port, opt_redir_get));
static       rxx  redirect_url  (sredirect_url.cstr ());

static const str shttp_url      (strbuf ("^%s%s%s%s$", opt_http, 
					 host, opt_port, opt_trailer));
static       rxx  http_url      (shttp_url.cstr ());

static       rxx  nums   ("^([0-9]+)$");
static       rxx  ipnums ("^([0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+)(\\.([0-9]+))?$");

static int
resolve_uri_full_parse_host (const char *host, taddr_t &home)
{
  //    www.x.com[.80]  
  char *loc = strrchr (host, *("."));

  if (loc) {
    size_t prfx = loc - host;
    if (prfx >= strlen (host))
      return -1;

    str num (loc+1);
    if (nums.match (num)) {
      if (ipnums.match (host)) {
	// 127.0.0.1[.80]
	home.set_name (ipnums[1]);
	home.port = ipnums[3] ? atoi (ipnums[3]) : http_port;
      }
      else {
	// www.x.com[.80]
	str name (host, prfx);      
	home.set_name (name);
	home.port = atoi (num);
      }
      return 0;
    }
  }
  if (!strcmp (host, redir_prfx))
    // redirect.nyud.net
    return -1;

  home.set_name (host);
  home.port = http_port;
  return 0;
}



/* 
 * Expected string:
 *    http://www.x.com[.80].nyud.net:8090/index.html
 *
 *    coral.name   = www.x.com[.80].nyud.net
 *    coral.port   = 8090
 *    coral.suffix = index.html
 *
 *    home.name    = www.x.com
 *    home.port    = 80
 *    home.suffix  = index.html
 */
static int 
resolve_uri_full (target_t &uri, std::string &path)
{
  taddr_t &coral = uri.coral_uri;
  taddr_t &home  = uri.home_uri;
  char *slash = "/";

  
    str chost;
    
    chost = path.c_str();
    
    str name;

    // example.com.nyud.net
    if (hostname_full.match (chost)) {
      if (hostname_full[1] == cdn_prfx)
	return -1;

      name = strbuf () << hostname_full[1] << cdn_str;
      coral.port = (hostname_full[3]) ? atoi (hostname_full[3]) : 80;

      if (resolve_uri_full_parse_host (hostname_full[1], home) < 0)
	return -1;
    }
    // example.com.http.l2.l1.l0.nyucd.net
    else if (hostname2_full.match (chost)) {
      if (hostname2_full[1] == cdn_prfx)
	return -1;

      // Grab prefix before opt_port
      if (!hostname2_full[4]) {
	coral.port = http_port;
	name = chost;
      }
      else {
	coral.port = atoi (hostname2_full[4]);

	// Grab contents before ":"
	const char *cchost = chost.cstr ();
	char *loc = index (cchost, ':');

	assert (loc);
	ssize_t len = loc - cchost;
	assert (len > 0 && (size_t) len < chost.len ());
	name = str (cchost, len);
      }
      
      if (resolve_uri_full_parse_host (hostname2_full[1], home) < 0)
	return -1;
    }
    else
      return -1;

    assert (name);
    coral.set_name (name);


  home.suffix = coral.suffix;
  return 0;
}
